/*
 *			W A P I. C
 * Manages wireless API to wireless interface drivers.
 */

#include "vbrsr.h"

#define SCAN_REQUEST 50

char *ifname;				/* interface name */

static int skfd = -10;			/* generic raw socket descriptor */

static bool werate;			/* can read bit rate through WE */
static bool weretry;			/* can read max retry number through WE */
static bool werange;			/* driver capabilities are valid */
static bool wescan;			/* driver has scanning capability */
static bool westats;			/* can read quality statistics via WE */
static int channel = -1;		/* frequency channel */
static iwrange range;			/* ranges for driver capabilities */


/*
 * Initialise the necessary structures to access the wireless driver
 * parameters, possibly using different methods.
 * The Wireless Extension API is deduced from the iwconfig and iwlist sources.
 *
 * input: the interface name (ifname)
 * output: a series of flags indicating the available capabilities
 */
void
init_wireless_api(void)
{
  struct iwreq wrq;			/* wireless interface request */
  iwstats stats;			/* quality statistics */

  assert(skfd == -10);			/* first time called */
  assert(ifname != NULL);		/* ifname initialised */

  /* Try to open the Wireless Extensions. */
  skfd = iw_sockets_open();
  werange = (skfd >= 0
	     && iw_get_range_info(skfd, ifname, &range) >= 0);

  /* Get channel through the Wireless Extensions. */
  if(iw_get_ext(skfd, ifname, SIOCGIWFREQ, &wrq) >= 0)
    channel = iw_freq_to_channel(iw_freq2float(&wrq.u.freq), &range);

  /* Get bit rate through the Wireless Extensions. */
  werate = (skfd >= 0
	    && iw_get_ext(skfd, ifname, SIOCGIWRATE, &wrq) >= 0);

  /* Get max retry limit through the Wireless Extensions. */
  weretry = (skfd >= 0 && werange
	     && range.retry_flags & IW_RETRY_LIMIT
	     && !(range.retry_flags & IW_RETRY_MIN) /* limit is fixed */
	     && iw_get_ext(skfd, ifname, SIOCGIWRETRY, &wrq) >= 0);

  /* Get quality statistics through the Wireless Extensions. */
  wrq.u.data.pointer = (caddr_t)&stats;
  wrq.u.data.length = 0;
  wrq.u.data.flags = 1;			/* Clear updated flag */
  strncpy(wrq.ifr_name, ifname, IF_NAMESIZE);
  westats = (skfd >= 0
	     && iw_get_ext(skfd, ifname, SIOCGIWSTATS, &wrq) >= 0);

 /* Try scaninng mode. */
  wrq.u.data.pointer = NULL;
  wrq.u.data.length = 0;
  wrq.u.data.flags = 0;
  wescan = (skfd >= 0
	    && iw_set_ext(skfd, ifname, SIOCSIWSCAN, &wrq) >= 0);
}

/*
 * Return a pointer to the iwrange structure,
 * containing various working ranges for parameters.
 * return NULL on error
 */
iwrange *
get_ranges()
{
  if (werange)
    return &range;
  else
    return NULL;
}

/*
 * Return an integer indicating the channel.
 * return a negative number on error
 */
int
get_channel_number()
{
  return channel;
}

/*
 * Return an integer indicating bit rate in Mb/s.
 * return -1 on error
 */
int
get_bit_rate()
{
  struct iwreq wrq;			/* wireless interface request */

  if (werate
      && iw_get_ext(skfd, ifname, SIOCGIWRATE, &wrq) >= 0)
    return wrq.u.bitrate.value / 1000000;
  else
    return -1;
}

/*
 * Whether the bit rate is automatic.
 * return -1 on error
 */
int
get_auto_bit_rate()
{
  struct iwreq wrq;			/* wireless interface request */

  if (werate
      && iw_get_ext(skfd, ifname, SIOCGIWRATE, &wrq) >= 0)
    return !wrq.u.bitrate.fixed;
  else
    return -1;
}

/*
 * Get the max number of ARQ retransmissions
 * argument is IW_RETRY_MIN for the short retry limit (RTS/CTS case)
 * argument is IW_RETRY_MAX for the long retry limit (non RTS/CTS case)
 * return -1 on error
 */
int
get_retry_limit(int flag)
{
  struct iwreq wrq;			/* wireless interface request */
  wrq.u.retry.flags = flag;		/* short or long retry limit */

  if (weretry
      && (flag == IW_RETRY_MIN || flag == IW_RETRY_MAX)
      && iw_get_ext(skfd, ifname, SIOCGIWRETRY, &wrq) >= 0)
    return (wrq.u.retry.disabled) ? 0 : wrq.u.retry.value;
  else
    return -1;
}

/*
 * Return a pointer to the iwstat structure,
 * contaning quality levels and missed packets.
 * return NULL on error
 */
iwstats *
get_stats()
{
  struct iwreq wrq;			/* wireless interface request */
  static iwstats stats;			/* contents of /proc/net/wireless */

  if (westats)
    {
      wrq.u.data.pointer = (caddr_t)&stats;
      wrq.u.data.length = 0;
      wrq.u.data.flags = 1;			/* Clear updated flag */
      strncpy(wrq.ifr_name, ifname, IF_NAMESIZE);
      if (iw_get_ext(skfd, ifname, SIOCGIWSTATS, &wrq) >= 0)
	{
	  if (stats.qual.level == 0
	      || ((stats.qual.updated & (IW_QUAL_QUAL_INVALID | IW_QUAL_QUAL_UPDATED))
		  != IW_QUAL_QUAL_UPDATED))
	      stats.qual.qual = 0;
	  if (stats.qual.level == 0
	      || ((stats.qual.updated & (IW_QUAL_LEVEL_INVALID | IW_QUAL_LEVEL_UPDATED))
		  != IW_QUAL_LEVEL_UPDATED))
	      stats.qual.level = 0;
	  if (stats.qual.level == 0
	      || ((stats.qual.updated & (IW_QUAL_NOISE_INVALID | IW_QUAL_NOISE_UPDATED))
		  != IW_QUAL_NOISE_UPDATED))
	      stats.qual.noise = 0;
	  return &stats;
	}
    }
  return NULL;
}

/*
 * Print one element from the scanning results
 */
static inline int
print_scanning_token(struct iw_event *	event,	/* Extracted token */
		     int		ap_num,	/* AP number */
		     struct iw_range *	iw_range,	/* Range info */
		     int		has_range,
                     FILE* file)
{
  char		buffer[128];	/* Temporary buffer */

  /* Now, let's decode the event */
  switch(event->cmd)
    {
    case SIOCGIWAP:
#if 0
      fprintf(file,"\n# Cell %02d - Address: %s\n", ap_num,
	     iw_pr_ether(buffer, event->u.ap_addr.sa_data));
#endif
      ap_num++;
      break;
    case SIOCGIWNWID:
      if(event->u.nwid.disabled)
	fprintf(file,"                    NWID:off/any\n");
      else
	fprintf(file,"                    NWID:%X\n", event->u.nwid.value);
      break;
    case SIOCGIWFREQ:
      {
	double		freq;			/* Frequency/channel */
	int		channel = -1;		/* Converted to channel */
	freq = iw_freq2float(&(event->u.freq));
	/* Convert to channel if possible */
	if(has_range)
	  channel = iw_freq_to_channel(freq, iw_range);
	iw_print_freq(buffer, sizeof(buffer),
		      freq, channel, event->u.freq.flags);
	fprintf(file,"# %s\n", buffer);
      }
      break;
    case SIOCGIWMODE:
      fprintf(file,"# Mode:%s\n",
	     iw_operation_mode[event->u.mode]);
      break;
    case SIOCGIWNAME:
      fprintf(file,"                    Protocol:%-1.16s\n", event->u.name);
      break;
    case SIOCGIWESSID:
      {
	char essid[IW_ESSID_MAX_SIZE+1];
	if((event->u.essid.pointer) && (event->u.essid.length))
	  memcpy(essid, event->u.essid.pointer, event->u.essid.length);
	essid[event->u.essid.length] = '\0';
	if(event->u.essid.flags)
	  {
	    /* Does it have an ESSID index ? */
	    if((event->u.essid.flags & IW_ENCODE_INDEX) > 1)
	      fprintf(file,"                    ESSID:\"%s\" [%d]\n", essid,
		     (event->u.essid.flags & IW_ENCODE_INDEX));
	    else
	      fprintf(file,"# ESSID:\"%s\"\n", essid);
	  }
	else
	  fprintf(file,"                    ESSID:off/any\n");
      }
      break;
    case SIOCGIWENCODE:
      {
	unsigned char	key[IW_ENCODING_TOKEN_MAX];
	if(event->u.data.pointer)
	  memcpy(key, event->u.essid.pointer, event->u.data.length);
	else
	  event->u.data.flags |= IW_ENCODE_NOKEY;
	fprintf(file,"# Encryption key:");
	if(event->u.data.flags & IW_ENCODE_DISABLED)
	  fprintf(file,"off\n");
	else
	  {
	    /* Display the key */
	    iw_print_key(buffer, sizeof(buffer), key, event->u.data.length,
			 event->u.data.flags);
	    fprintf(file,"%s", buffer);

	    /* Other info... */
	    if((event->u.data.flags & IW_ENCODE_INDEX) > 1)
	      fprintf(file," [%d]", event->u.data.flags & IW_ENCODE_INDEX);
	    if(event->u.data.flags & IW_ENCODE_RESTRICTED)
	      fprintf(file,"   Security mode:restricted");
	    if(event->u.data.flags & IW_ENCODE_OPEN)
	      fprintf(file,"   Security mode:open");
	    printf("\n");
	  }
      }
      break;
    case SIOCGIWRATE:
      iw_print_bitrate(buffer, sizeof(buffer), event->u.bitrate.value);
      fprintf(file,"# Bit Rate:%s\n", buffer);
      break;
    case IWEVQUAL:
      {
	iw_print_stats(buffer, sizeof(buffer),
		       &event->u.qual, iw_range, has_range);
	fprintf(file,"# %s\n", buffer);
	break;
      }
    case IWEVCUSTOM:
      {
	char custom[IW_CUSTOM_MAX+1];
	if((event->u.data.pointer) && (event->u.data.length))
	  memcpy(custom, event->u.data.pointer, event->u.data.length);
	custom[event->u.data.length] = '\0';
	fprintf(file,"                    Extra:%s\n", custom);
      }
      break;
    default:
      fprintf(file,"                    (Unknown Wireless Token 0x%04X)\n",
	     event->cmd);
   }	/* switch(event->cmd) */

  /* May have changed */
  return(ap_num);
}

void
print_scan_info(FILE* file)
{
 struct iwreq wrq;
 unsigned char* buffer=NULL;              // ATTENZIOANE : Presi e non valutati
 int bufflen = IW_SCAN_MAX_DATA;          // da iwlist.c
 struct timeval tv;                       // Upper bound per la select

 tv.tv_sec=3;    // Vedere se e' sufficientemente lungo
 tv.tv_usec=0;

 if(wescan < 0)
 {
  printf("debug : print_scan_info() : L'interfaccia %s non supporta lo scanning.\n",ifname);
  return;
 }

 while(1)
 {
  fd_set rfds;        /* File descriptor per l'utilizzo della select per il controllo dei descrittori in lettura*/
  int ret;
  FD_ZERO(&rfds);
  ret=select(0,&rfds,NULL,NULL,&tv);
  /* Errore nella select */
  if(ret < 0)
  {
   if(errno==EAGAIN || errno==EINTR)
    continue;
   printf("debug : La select ha generato un errore\n");
   return;
  }
   /* Il timer e' spirato , controllo se ho un risultato oppure rientro nel ciclo*/
  if(ret == 0)
  {
   buffer=malloc(bufflen);
   wrq.u.data.pointer=buffer;
   wrq.u.data.flags=0;
   wrq.u.data.length=bufflen;
   if(iw_get_ext(skfd,ifname,SIOCGIWSCAN,&wrq) < 0)
   {
    if(errno==EAGAIN)
     continue;
    return;
   }
   /* Ho un risultato : verifico,altrimenti ritento
      1) Andrebbe valutato un meccanismo che fa piu' di una chiamata
         fino alla scoperta di una rete */
   else
    break;
  }
 }

 /* Codice estratto da iwlist.c*/
 if(wrq.u.data.length)
 {
  struct iw_event iwe;
  struct stream_descr stream;
  int ap_num = 1;
  int ret;
  struct iw_range range;
  int has_range;

  has_range=(iw_get_range_info(skfd,ifname,&range) >=0 );
  if((!has_range) || (range.we_version_compiled <14))
  {
    fprintf(stderr, "%-8.16s  Interface doesn't support scanning.\n\n",
	    ifname);
    return;
  }

  fprintf(file,"# %-8.16s  Scan completed :\n", ifname);

  iw_init_event_stream(&stream, buffer, wrq.u.data.length);
  do
  {
   ret = iw_extract_event_stream(&stream, &iwe,
				   range.we_version_compiled);
   if(ret > 0)
   {
     ap_num = print_scanning_token(&iwe, ap_num, &range, has_range,file);
   }
  }
  while(ret > 0);
 }
 else
   printf("%-8.16s  No scan results\n", ifname);
 free(buffer);
 printf("\n");
 return;
}



/*
 * Local Variables:
 * comment-column: 40
 * fill-column: 79
 * c-font-lock-extra-types: ("FILE"  "\\sw+_t" "bool" "iwrange" "iwstats")
 * End:
 */
